#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  5 04:26:07 2017

@author: Daniel Speckhard dts@stanford.edu

This package contains methods to transform sensor data into more
palatable forms for topology estimation.
"""
import redundant_data
import transform


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 04:39:06 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains methods relating to calculating the mutual information
between two distributions.
"""

import gaussian
import discrete
import jvhw



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 04:39:06 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains methods relating to calculating the mutual information
between two distributions using entropy estimators developed by the Weissman
Group at Stanford University.
"""

import est_MI as jvhw
import numpy as np

def find_jvhw_mi(vmag_matrix):

    """ Finds the mutual information of node pairs with JVHW estimators. 
        
    This function uses the JVHW entropy estimators to cacluate mutual
    information between nodes. This function acts as a shell to call
    a script written by the Weissman group to calculate mutual information
    between a pair of nodes. 
        
    Args
    ----------
    vmag_matrix: 2D ndarray, shape (# of data-ponts, # of nodes)
        The matrix contains the voltage magntiude or the change in 
        voltage magntiude (if the method take_deriv() has been called).
            
    Returns
    ----------
    mi_matrix: 2D ndarray, shape (# of nodes, # of nodes)
        The matrix will be non-zero only for lower triangular values 
        since MI is symmetric and self-information values are of no use
        to us.  The matrix will be calculated using JVHW estimators for
        entropy.
    """
    num_buses = len(vmag_matrix[1,:])
    mi_matrix = np.zeros((num_buses, num_buses))
    for i in range(1, num_buses):
        for j in range(0, i):
            mi_matrix[i, j] = jvhw.est_MI_JVHW(
                    vmag_matrix[:, i], vmag_matrix[:, j])
    return mi_matrix


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 04:39:06 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains methods relating to calculating the mutual information
between two distributions using the discrete mutual information formula. There
are two methods here. 1) Caclulate the discrete mutual 
information using methods from SK.learn. 2) Use maximum
likelihood estimators for mutual infromation written by the Weissman group at
Stanford, which in effect calculates the mutual information using the 
well-known discrete formula. 
"""

from sklearn.metrics import mutual_info_score
import est_MI as jvhw
import numpy as np

def find_sk_discrete_mi(vmag_matrix, bins):
        
    """ Finds the mutual information using discrete formula.
        
    This function calculates the mutual information between all the sets of
    two distinct nodes using the discrete formula for mutual information. 
        
    Args
    ----------
    vmag_matrix: ndarray, shape (# of data-ponts, # of nodes)
        The matrix contains the voltage magntiude data or the
        change in voltage magntiude data (if take_deriv() method has 
        been called).
            
    bins: int
        This controls how many bins to bin the continuous data
        from self.vmag_matrix. We must discretize the data before 
        taking the discrete mutual information. This parameter will 
        control the computation time of this function as well as the 
        precision.
    
    Returns
    ----------
    mi_matrix: 2D ndarray, shape (# of nodes, # of nodes)
        The matrix will be non-zero only for lower triangular values 
        since MI is symmetric and self-information values are of no use
        to us. The matrix will be calculated from the formula Sum_ij 
        p(i,j)log(p(i,j)/(p(i)*p(j))). """
            
    num_buses = len(vmag_matrix[1,:])
    mi_matrix = np.zeros((num_buses, num_buses))
    for i in range(1, num_buses):
        for j in range(0, i):
            mi_matrix[i, j] = _calc_discrete_mi(
                    vmag_matrix[:, i], vmag_matrix[:, j], bins)
    return mi_matrix

def _calc_discrete_mi(first_vec, second_vec, bins):
    """ Calculates discrete mutual information between two nodes.
            
    This function is a shell function to call sklearn's mutual info
    score function to calculate the mutual information between one pair
    of nodes. 
            
    Args
    ----------
    first_vec: 1D ndarray, shape (# of datapoints)
        This is the vector of voltage magnitude data for one
        node.
                
    second_vec: 1D ndarray, shape (# of datapoints)
        This is the vector of voltage magntiude data for 
        the second node. Both the first and second node form the 
        pair of nodes for which we calculate the mutual 
        information. Both vectors should be the same size.
                
    bins: int
        This controls how many bins to bin the ~continuous 
        voltage magnitude data. More bins means more precision in 
        the discretization. 
                
    Returns
    ----------
    mutual_info: double
        The mutual information, in bits, between the 'first'
        and 'second' node calculated using the discrete mutual
        information formula. 
    """
    c_xy = np.histogram2d(first_vec, second_vec, bins)[0]
    mutual_info = mutual_info_score(None, None, contingency=c_xy)
    return mutual_info

def find_mle_mi(vmag_matrix):
    """ Finds the MLE mutual information of node pairs. 
        
    This function uses MLE entropy estimators to cacluate mutual
    information between nodes. This function acts as a shell to call
    a script written by the Weissman group to calculate mutual information
    between a pair of nodes. 
        
    Args
    ----------
    vmag_matrix: 2D ndarray, shape (# of data-ponts, # of nodes)
        The matrix contains the voltage magntiude or the change in 
        voltage magntiude (if the method take_deriv() has been called).
            
    Returns
    ----------
    mi_matrix: 2D ndarray, shape (# of nodes, # of nodes)
        The matrix will be non-zero only for lower triangular values 
        since MI is symmetric and self-information values are of no use
        to us.  The matrix will be calculated using JVHW estimators for
        entropy.
    """
    num_buses = len(vmag_matrix[1,:])
    mi_matrix = np.zeros((num_buses, num_buses))
    for i in range(1, num_buses):
        for j in range(0, i):
            mi_matrix[i, j] = jvhw.est_MI_MLE(
                    vmag_matrix[:, i], vmag_matrix[:, j])
    return mi_matrix


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mar 26 05:51:21 2017

@author: Daniel Speckhard, dts@stanford.edu

This module contains methods to transform sensor data to improve
the efficiency or accuracy of the grid estimation algorithm when
run on the resulting transformed data.
"""
import numpy as np

def discretize_signal(vmag_matrix, bits):

    """ 
    Discretizes data with an input number of bits precision. 
        
    This function discretizes the voltage magnitude data. 
    self.vmag_matrix: A matrix containing the voltage magnitude data.
    The number of columns should be the number of nodes in the system. The
    number of rows in the matrix corresponds to the number of datapoints
    where the voltage magnitude is measured across all nodes.
        
    Args
    -------
    bits: int
        Exponential number of bins with which to discreteize data. 
        Meaning, if bins = 10, we bin the data into 2^10 bins. In this 
        scheme we make each bin equal size.
            
    Returns
    -------
        vmag_matrix: 2D ndarray, shape (# of data-points, # of nodes)
        This matrix is the size of the input
        vmag_matrix. The values in this matrix are all integers. 
        The values in the output version of the matrix are discretized 
        are equal to the bin number which the value of the input matrix
        falls into with respect to the binning process.
    """
    global_min = np.min(np.min(vmag_matrix))

    global_max = np.max(np.max(vmag_matrix))

    # Shift the data so new minimum is at zero.
    vmag_matrix = vmag_matrix - global_min
    # Determine the bin-size for equally spaced bins.
    bin_size = np.divide((global_max-global_min), (2**bits - 1))
    # Now find new values for data:
    vmag_matrix = \
        np.round(np.divide(vmag_matrix, bin_size)).astype(int) 
    return vmag_matrix
    
def get_delta(vmag_matrix, deriv_step):
        
    """ Convert voltage mag. data into the change in voltage mag. wrt time.  
        
    This function finds the change in voltage magntiude between 
    time-points. The spacing between time-points is a variable that can be
    changed. The change in voltage magnitude is calculated as Delta_Vmag =
    Vmag(node x, time t) - Vmag(node x, time t-deriv_step).
        
    Args
    -------
    vmag_matrix: 2D ndarray, shape (# of data-points, # of nodes)
        This is the voltage magnitude data at different nodes. We want to 
        transform this data so we end up with the change
        in voltage magnitude from one time-point to the next for each node.
        
    deriv_step: int 
        This parameter decides how many time-points for which
        to compute the voltage magntiude difference at a node. Ex. if
        deriv_step = 5, then the change in voltage magnitude will be
        Delta_Vmag = Vmag(node x, time t) - Vmag(node x, time t-5).
        
    Returns
    -------
    vmag_matrix: 2D ndarray, shape (# of data-points, # of nodes)
        This matrix now represents the change in voltage
        magnitude at a node from one time-point to the next. Each column
        corresponds to a different node and each row to a time-point.
    """
    end = len(vmag_matrix[:, 1])-1
    vmag_matrix = vmag_matrix[
            deriv_step+1:-1, :] - vmag_matrix[1:end-deriv_step, :]
    
    return vmag_matrix


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mar 25 02:32:07 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains a methods to clean input voltage magntiude data and the 
true branch list of the network thereafter.
"""

import numpy as np

def collapse_redundant_data(vmag_matrix, true_branches):

    """
    Removes nodes in the network with identical data to other nodes.

    This function will check to see if there are any nodes that contain
    voltage measurements that are the same as another node. For instance,
    if 40 nodes are given, this function will check if Node 1 has the same
    data as Node 2. If Node 1 and Node 2 have the same data this
    function will collapse the two nodes, meaning it will remove Node 2
    and fix the list of true branches to replace any reference to Node 2
    with Node 1.

    Args
    -------
    vmag_matrix: 2D ndarray, shape (# of data-points, # of nodes)
        A matrix containing the voltage magnitude data.
        The number of columns should be the number of nodes in the system.
        The number of rows in the matrix corresponds to the number of
        datapoints where the voltage magnitude is measured across all
        nodes. When this function is called the matrix should not be empty.

    true_branches: 2D ndarray, shape (# of nodes - 1, 2)
        This is a matrix containing the true branches in
        the system. The number of rows in this matrix corresponds to the
        number of true branches in the system. The number of columns should
        be equal to two, which is the number of nodes involved in a single
        branch.

    Returns
    -------
    vmag_matrix: 2D ndarray, shape (# of data-points,
                                             # of unique nodes)
        This matrix is the voltage magnitude dataset with the redundant
        nodes removed. Therefore, the output matrix will be different size
        than the input matrix if redundant nodes are present in the input
        matrix. The word redundant here refers to a node that has
        all the same data values as another node. The node which is lower
        in number is retained when a redundant node is found. Ex. If Node
        34 has the same data as Node 1. Node 34 will be removed from node
        volt matrix.

    true_branches: 2D ndarray, shape (# of nodes - 1, 2)
        The list of true branches is also an output, it
        has redundant nodes replaced with lower node label number
        non-redundant nodes. Ex. Imagine Node 34 has the same data as Node
        1. Any branches in the true_branches matrix containing node 34,
        let's say a branch (7,34). The output true_branches will now list
        (7,1). If there's a branch from (1,34) it will have to be deleted
        since a branch (1,1) is a self-referring branch. Self-referring
        branches are sometimes returned by this function. This function
        does not remove self-referring branches.
    """
    num_buses = len(vmag_matrix[1,:])
    number_observations = len(vmag_matrix[:, 1])

    # We initialize a variable for the number of identical nodes found.
    # This is useful for debugging purposes.
    number_identical_nodes = 0

    # We will cycle through the node_volt_matrix and see if we can find
    # redundant nodes. If so we will delete this node. Note the number of
    # columns (i.e. nodes) is changing, this means we need to
    # evaluate this number every loop of the cycle.

    # We have to use a while loop since in Matlab for loops evaluate
    # bounds at first run and we will change the size of our matrix that
    # we iterate through, since we find identical pairs, we'll remove
    # columns.
    outer_loop_index = 0

    matrix_of_identical_nodes = np.zeros(((num_buses -1), 2))

    while outer_loop_index < (len(vmag_matrix[1, :]) -1):
        # We want to evaluate whether one node is the same as another
        # column so we check starting with node i+1.
        inner_loop_index = outer_loop_index + 1
        while inner_loop_index <= (len(vmag_matrix[1, :]) -1):
            # Check whether it is the same node.
            if np.sum(vmag_matrix[:, outer_loop_index] ==
                      vmag_matrix[:, inner_loop_index]) == \
                      number_observations:

                number_identical_nodes = number_identical_nodes + 1
                matrix_of_identical_nodes[number_identical_nodes, :] = \
                                             outer_loop_index, inner_loop_index
                # Now let's remove the identical nodes.
                vmag_matrix = \
                np.delete(vmag_matrix, inner_loop_index, 1)

                # Update the list of true branches.
                # Go through the rows of the true branch data set.
                for i in range(0, num_buses -1):
                # Check if we find the second identical node-k in the row.
                    for j in range(0, 2):
                        if true_branches[i, j] == inner_loop_index:
                            # set the higher number redundant node equal to the 
                            # lower number redundant node.
                            true_branches[i, j] = outer_loop_index
                        elif true_branches[i, j] > inner_loop_index:
                    # Since I've just deleted a col, I also need to
                    # downshift all the values in the True Branch Data set
                    # that were above this col. Ex. if i delete column 5
                    # (node 5), i need to make nodes 6 appear as node 5 in
                    # the true branch data.
                            true_branches[i, j] = \
                                         true_branches[i, j] -1

        # If a column was not deleted continue iterating. We don't
        # increment iterator if a col was deleted since now col k is a
        # different col.
            else:
                inner_loop_index = inner_loop_index +1
        outer_loop_index = outer_loop_index + 1

    # Update self.num_buses
    num_buses = len(vmag_matrix[1, :])
    return vmag_matrix, true_branches

def remove_redundant_branches(true_branches):
    """
    Removes branches that start at one node and end at the same node.

    This function removes self-referring branches from true_branches.
    Ex. if there is a branch in self.true_branches such as (1,1), it will
    be removed. Redundant branches (like (1,1)) should exist in
    self.true_branches only after the method collapse_redundant_nodes
    has been called.

    Args
    -------
    true_branches: 2D ndarray, shape (# of nodes - 1, 2)
        This is a matrix containing the true branches in
        the system. The method remove_redundant_branches is called when we
        suspect self.true_branches has additional self-referring branches.
        The number of rows in this matrix corresponds to the
        number of true branches in the system plus any self-referring
        branches. The number of columns should be equal to two, which is
        the number of nodes involved in a single branch.

    Returns
    -------
    true_branches: 2D ndarray, shape (# of nodes - 1, 2)
        This is a matrix containing the true branches in
        the system. The number of rows in this matrix corresponds to the
        number of true branches in the system since this method will have
        removed any self-referring (redundant) branches. The number of
        columns should be equal to two, which is the number of nodes
        involved in a single branch.
    """

    branch_counter = 0;

    # Check every branch.
    while branch_counter < len(true_branches[:,1]):
        # Check if both nodes in a single branch are equal to each other.
        if true_branches[branch_counter, 0] == \
                        true_branches[branch_counter,1]:
            # delete the row containing the self-referrring branch.
            np.delete(true_branches, branch_counter, 0)
        else: # increment counter so we can check the next branch.
            branch_counter = branch_counter + 1;

    return true_branches


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  7 22:21:28 2017

@author: Daniel Speckhard, DTS@stanford.edu

This library was created at the Stanford Sustainable Systems Lab. This library
includes functions to estimate the topology of an electrical grid given voltage
measurements at nodes.
"""
import numpy as np
import networkx as nx
import mi_alg
import matplotlib.pyplot as plt
import transform_data
import mutual_information

class GridEst(object):


    """ Class to statistically estimate topology of electrical grid.
    
    GridEst is a class containing variables and methods to estimate
    the topology of system using a variety of statistical methods from voltage
    magnitude data taken at each node in the network.
    
    Currently, only the Chow-Liu algorithm has been implemented. In the future
    a LASSO method will be implemented. 
    """
    
    def __init__(self, true_branches, vmag_matrix, graph_name, num_bits):

        """ Initialize grid est class instance.

        Args
        ----------
        true_branches: ndarray, shape (# of nodes - 1, 2)
            This is matrix storing the node lables of each pair of nodes
            connected by a true branch in the true tree structure. Ex. [(2,1),
            (3,1)] would indicate that nodes 2 and 3 are both connected to
            Node 1.

        vmag_matrix: ndarray, shape (# of datapoints, # of nodes)
            This matrix should contain the voltage magnitude
            data at different timepoints and different nodes. The matrix should
            be size ((number of nodes) x (number of data points at each node)).
            
        graph_name: str
            This is the name of the network. When we plot the estimated newtork
            we will save a figure with this name. 

        Returns
        ----------
        self.num_buses: int
            This is the number of buses in the network.

        self.mi_matrix: str
            This variable is initialized to a string labeled undefined so that
            the user is sure that the values have not yet been calculated. When
            the values are calculated self.mi_matrix should become an ndarray
            shape (# of nodes, # of nodes).

        self.est_branches: ndarray, shape (# of nodes - 1, 2)
            These are the branches that are estimated by the algorithm. The
            values are intialized to zero (bogus values) and should be filled
            by calling appropriate methods.

        self.graph: str
            This variable should eventually store the Networkx object graph
            ouput from estimation. The variable is initialized to 'undefined'
            so that the user knows the estimation has not yet taken place.
        
        self.graph_name: str
            The name of the graph which will be estimated. 
        
        self.num_bits: int
            The number of bits to which to discretize the input data when 
            JVHW or discrete mutual information methods are used for analysis.
            When the Gaussian method is used for analysis the data is not
            discretized and this numer is not used. The number of bits of
            precision specifies the number of equally spaced bins with which
            to bin the non-discretized data. Ex. num_bits = 2: then 2^2 bins
            of equal space will be used to bin the data into 2^2 bins and
            the discretized data will take on values from 0 to (2^2 - 1).

        """
        self.true_branches = true_branches -1
        self.vmag_matrix = vmag_matrix
        self.mi_matrix = 'undefined' # Leave this undefined until later.
        self.num_buses = len(self.vmag_matrix[0, :])
        self.est_branches = np.zeros((self.num_buses - 1, 2))
        self.graph = 'undefined'
        self.graph_name = graph_name
        self.num_bits = num_bits


    def find_sdr(self):

        """ Find the succesfull detection rate of estimating branches.

        This function returns the percentage of estimated branches which are
        true branches (branches contained in self.true_branches).

        Args
        ----------
        self.est_branches: 2D ndarray, shape (number of nodes - 1, 2)
            Each row is a pair of nodes corresponding to an estimated branch
            from the algorithm output. The number of branches should be one
            less than the total number of nodes since we are estimating a tree.

        self.true_branches: 2D ndarray, shape (number of nodes - 1, 2)
            This is a matrix containing the true branches in the system.
            The number of rows in this matrix corresponds to the number of
            true branches in the system. The number of columns should be equal
            to two, which is the number of nodes involved in a single branch.

        Returns
        ----------
        successful_branch_counter: int
            This is a integer counter that tells us how many branches have been
            correctly estimated by the algorithm.

        sdr: double
            This value is the successful detection rate (SDR) percentage of
            the estimation algorithm. It equals the number of correctly
            identified branches divided by the total number of branches.
        """

        successful_branch_counter = 0.0
        # Update self.est_branches to be 1 larger since we've renumered the
        # other nodes.
        #self.est_branches = self.est_branches + 1
        #flipped_est_branches = np.fliplr(self.est_branches)
        for i in range(0, np.shape(self.est_branches)[0]):
            if ([self.est_branches[i, 0], self.est_branches[i, 1]] \
            in self.true_branches.tolist()) or ([self.est_branches[i, 1],
                                                 self.est_branches[i, 0]] in
                                                self.true_branches.tolist()):

                successful_branch_counter += 1.

        sdr = np.divide(successful_branch_counter,
                        len(self.est_branches[:, 0]))
        # Multiply by 100 to get the percent
        sdr = 100.0*sdr

        return sdr, successful_branch_counter

    def run_mi_alg(self, mi_method):

        """ This method estimates the topology of network using Chow-Liu Alg.

        The chow-liu algorithm is run wherby a network is estimated using the
        minimum spanning tree with the mutual information between nodes used
        as weights. The input data is voltage magntiude data taken at
        each node at constant time intervals. The algorithm is run on the
        change in voltage magntiude between one time-point and the next for
        improved results.

        Args
        ----------
        self.vmag_matrix: ndarray, shape (# of datapoints, # of nodes)
            This matrix should contain the voltage magnitude
            data at different timepoints and different nodes.

        self.true_branches: 2D ndarray, shape (number of nodes - 1, 2)
            This is a matrix containing the true branches in the system.
            The number of rows in this matrix corresponds to the number of
            true branches in the system. The number of columns should be equal
            to two, which is the number of nodes involved in a single branch.

        self.graph: str
            This variable should eventually store the Networkx object graph
            ouput from estimation. The variable is initialized to 'undefined'
            so that the user knows the estimation has not yet taken place.
            
        self.num_bits: int
            The number of bits to which to discretize the input data when 
            JVHW or discrete mutual information methods are used for analysis.
            When the Gaussian method is used for analysis the data is not
            discretized and this numer is not used. The number of bits of
            precision specifies the number of equally spaced bins with which
            to bin the non-discretized data. Ex. self.num_bits = 2: then 2^2 
            bins of equal space will be used to bin the data into 2^2 bins and
            the discretized data will take on values from 0 to (2^2 - 1).
            
            
        mi_method: str
            This variable determines how to calculate the mutual information 
            between nodes. We can select between gaussian, sk_discrete, MLE or
            JVHW. The MLE and sk_discrete method should return the same values
            as both use the discrete mutual information formula to calculate 
            the mutual information, however, the implementations are different
            so there may be small differences. The gaussian method approximates
            the voltage magntiude data as a gaussian function. The JVHW method
            uses entropy estimators developed by the Weissman group at
            Stanford.

        Returns
        ----------
        A figure of the estimated graph is saved to the current directory as a
        PNG file.

        """
        # Remove nodes (if any) which contain redundant data.
        self.vmag_matrix, self.true_branches  = \
        transform_data.redundant_data.collapse_redundant_data(
                self.vmag_matrix,self.true_branches)
        # Clean the list of true branches after collapse redundant data
        # has been called. This removes false redundant branches (ex. (1,1))
        # which are a result of the above function call.
        self.true_branches = \
        transform_data.redundant_data.remove_redundant_branches(
                self.true_branches)
        
        # Take delta/deriv of data, with step-size/spacing of one data-point.
        deriv_step = 1
        self.vmag_matrix = transform_data.transform.get_delta(
                self.vmag_matrix, deriv_step)
        
        # Find the mutual information. First specify the # of bins to use when
        # JVHW or discrete MI methods are used.
        bins = 2**self.num_bits

        if mi_method == 'gaussian':
            entropy_vec = mutual_information.gaussian.find_gaussian_entropy(
                    self.vmag_matrix)
            joint_entropy_matrix = \
            mutual_information.gaussian.find_joint_gaussian_entropy(
                    self.vmag_matrix)
            self.mi_matrix = mutual_information.gaussian.find_mi_mat(
                    entropy_vec, joint_entropy_matrix)

        elif mi_method == 'sk_discrete':
            self.mi_matrix = mutual_information.discrete.find_sk_discrete_mi(
                    self.vmag_matrix, bins)
        
        elif mi_method == 'MLE':
            self.vmag_matrix = transform_data.transform.discretize_signal(
                    self.vmag_matrix, self.num_bits)
            self.mi_matrix = mutual_information.discrete.find_mle_mi(
                    self.vmag_matrix)

        elif mi_method == 'JVHW':
            self.vmag_matrix = transform_data.transform.discretize_signal(
                    self.vmag_matrix, self.num_bits)
            self.mi_matrix = mutual_information.jvhw.find_jvhw_mi(
                    self.vmag_matrix)
            
        else:
            print('Please select a valid mutual information method' + 
                  ' gaussian, sk_discrete, MLE or JVHW')
            raise SystemExit
        
        # Find the minimum spanning tree using mutual information as weights.
        self.est_branches, self.graph = mi_alg.find_mst(self.mi_matrix)
        
        # Plot the graph with 'twopi' graphviz layout.
        self.plot_graph('twopi')



    def plot_graph(self, graphviz_layout):
        """ Method to save newtorkx graph object as a dot file.

        This method plots the graph in self.graph using a graphviz layout
        program. The plot is saved as a png. This method is mostly a shell 
        function to call nx.draw_graphviz() to draw the newtorkx object.

        Args
        ----------
        self.graph: networkx graph object
            This object contains the estimated graph from the estimatio
            algorithm.
        
        self.graph_name: str
            This object will save the plotted estimated graph as a PNG file
            with this name. 

        graphviz_layout: string
            Speficied the layout of the graph when draw the graph onto a
            figure. We can choose from twopi, gvcolor, wc, ccomps,
            tred, sccmap, fdp, circo, neato, acyclic, nop, gvpr, dot, sfdp.
            
        Returns
        ----------
        The graph in self.graph is drawn onto a matlabplot.pyplot figure and
        saved as a PNG file. 
        """

        # One alternative option to this method is to save the graph as a
        # dot file with: nx.write_dot(self.graph,"grid.dot")
        
        self.renumber_graph()
        plt.figure()
        nx.draw_graphviz(self.graph, graphviz_layout, with_labels = True,
                         node_color = 'r', node_size = 200, font_size = 10 )

        plt.savefig(self.graph_name + '.png', format='PNG')
        plt.show()

    def renumber_graph(self):
        """ Renumber the graph so that the lowest numbered Node is Node 1.
        
        This method relables the nodes so that the node numbering begins
        from 1 rather than from 0.
        
        Args:
        ----------
        self.graph: networkx graph object
            This object contains the estimated graph from the estimation
            algorithm. The labelling here begins from 0.
        Returns:
        ----------
        self.graph: networkx graph object
            This object contains the estimated graph from the estimation
            algorithm. The labelling here bgins from 1.
        """
        mapping = {}
        
        for i in range(0, self.num_buses):
            mapping[i] = i + 1
                   
        self.graph = nx.relabel_nodes(self.graph, mapping)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

Created on Tue Mar  7 22:21:28 2017

@author: Daniel Speckhard, DTS@stanford.edu

This package includes functions to estimate the topology of an electrical grid 
given voltage measurements at nodes using different estimation techniques.
"""




import numpy as np
from est_entro import est_entro_JVHW, est_entro_MLE, formalize_sample

def est_MI_JVHW(X, Y):
    """This function returns our scalar estimate of mutual information I(X;Y)
    when both X and Y are vectors, and returns a row vector consisting
    of the estimate of mutual information between each corresponding column
    of X and Y when they are matrices.
    Input:
    ----- X, Y: two vectors or matrices with the same size, which can only
                contain integers.
    Output:
    ----- est: the estimate of the mutual information between input vectors
               or that between each corresponding column of the input
               matrices. The output data type is double.
    """
    [X, Y, XY] = formalize(X, Y)

    # I(X,Y) = H(X) + H(Y) - H(X,Y)
    return np.maximum(0, est_entro_JVHW(X) + est_entro_JVHW(Y) - est_entro_JVHW(XY))

def est_MI_MLE(X, Y):
    """This function returns the scalar MLE of the mutual information I(X;Y)
    when both X and Y are vectors, and returns a row vector consisting
    of the estimate of mutual information between each corresponding column
    of X and Y when they are matrices.
    Input:
    ----- X, Y: two vectors or matrices with the same size, which can only
                contain integers.
    Output:
    ----- est: the estimate of the mutual information (in bits) between input
               vectors or that between each corresponding column of the input
               matrices. The output data type is double.
    """

    [X, Y, XY] = formalize(X, Y)

    # I(X,Y) = H(X) + H(Y) - H(X,Y)
    return np.maximum(0, est_entro_MLE(X) + est_entro_MLE(Y) - est_entro_MLE(XY))


def formalize(X, Y):
    X = formalize_sample(X)
    Y = formalize_sample(Y)

    if X.shape != Y.shape:
        raise ValueError('Input arguments X and Y should be of the same size.')

    X = X.astype(np.int64, copy=False)
    Y = Y.astype(np.int64, copy=False)

    X = map_int(X)
    Y = map_int(Y)
    XY = (X - 1) * Y.max(axis=0) + Y

    return X, Y, XY

def map_int(samp):
    """Map integer data along each column of X and Y to consecutive integer
    numbers (which start with 1 and end with the total number of distinct
    values in each corresponding column). For example,
                    [  1    6    4  ]        [ 1  3  3 ]
                    [  2    6    3  ] -----> [ 2  3  2 ]
                    [  3    2    2  ]        [ 3  1  1 ]
                    [ 1e5   3   100 ]        [ 4  2  4 ]
    The purpose of this data mapping is to make the effective data range
    as small as possible, minimizing the possibility of overflows.
    """
    id = samp.argsort(axis=0)
    col_index = np.indices(samp.shape)[1]
    samp = samp[id, col_index]
    samp[id, col_index] = np.cumsum(np.r_[np.ones((1, samp.shape[1])), np.diff(samp, axis=0) > 0], axis=0)
    return samp


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

Created on Tue Mar  7 22:21:28 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module includes functions to estimate the topology of an electrical grid 
given voltage measurements at nodes using mutual information based
estimation.
"""

import networkx as nx
import numpy as np

def find_mst(mi_matrix):
        
    """ Finds the maximum spanning tree based on mutual information.
        
    This function finds the maximum spanning tree for the network using
    the mutual information values between nodes as weights. NetworkX's
    minimum spanning tree funciton is called on the recriprocal of the
    mutual information network, therefore generating the maximum
    spanning tree of the network.
      
    Args
    -------
    mi_matrix: 2D ndarray, shape (# of nodes, # of nodes)
        The mutual information matrix containing pair-wise
        mutual information for different nodes in the system. 
            
    Returns
    -------
    est_branches: 2D ndarray, shape (# of nodes - 1, 2)
        The estimated branches by running the minimum spaning
        tree algorithm using the mutual information values between 
        nodes as weights.
            
    graph: NetworkX graph object
        This is used for plotting/drawing the estimated graph. 
        """

    # We create a fully connected graph with weights specified by the
    # the mutual information values.
    # However, we have to take the reciprical of the MI matrix values
    # since nx only has min_span_tree as a fxn for python 2.7. We
    # add a tiny number to each value so reciprocal is defined.
    mi_matrix = mi_matrix + 1e-12
    # We cretae a temporary nx. graph object that's a cyclical graph
    # with edge weights equal to MI values in the matrix.
    net = nx.from_numpy_matrix(np.reciprocal(mi_matrix),
                               create_using=nx.Graph())
    mst = nx.minimum_spanning_tree(net)
    est_branches = np.array(mst.edges())
    graph = mst
    
    return est_branches, graph


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

Created on Tue Mar  7 22:21:28 2017

@author: Daniel Speckhard, DTS@stanford.edu

This package includes exmaples to estimate the grid topology using statistical
methods. 
"""



from distutils.core import setup
setup(
  name = 'grid_top_est',
  packages = ['estimation', 'examples', 'mutual_information', 'transform_data'],
  version = '0.2',
  description = 'Library to estimate topology of electrical grid using voltage sensor data',
  author = 'Daniel Speckhard',
  author_email = 'dts@stanford.edu',
  url = 'https://github.com/speckhard/Python_Grid_Est', # use the URL to the github repo
  download_url = 'https://github.com/speckhard/Python_Grid_Est/archive/0.2.tar.gz', 
  keywords = ['information-theory', 'smart-grid', 'estimation', 'toplogy', 'mutual-information'], 
  classifiers = [],
)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 08:43:21 2017

@author: Daniel Speckhard dts@stanford.edu

This script acts as an example to run the mutual information based grid
estimation aglorithm developed in the Sustainable Systems Lab at Stanford
University on a 8 bus IEEE network.
"""
import numpy as np
from scipy.io import loadmat
# This allows us to run the example as a main file and not as a package.
# It takes care of relative importing in python. 
if __name__ == '__main__' and __package__ is None:
    from os import sys, path
    sys.path.append(path.dirname(path.dirname(path.abspath(__file__)))) 
from estimation.grid_est import GridEst
# 8 Node Example.
matfile = loadmat('Node8_randPF_solar.mat',
                  squeeze_me=True, struct_as_record=False)
# Load the voltage magnitude matrix.
vmag_matrix = matfile['v_vec']
# Remove the feeder node data from v_vec.
vmag_matrix = np.delete(vmag_matrix, 0, 1)
# Load the ture branches of the network.
mpc_base = matfile['mpc_base']
# Take the columns corresponding to true branches.
# Subtract by 1, to account for feeder node being deleted.
true_branches = mpc_base.branch[:,0:2]
# Remove the feeder branch (1-2) from the true branches array.
# Subtract by 1, to account for feeder node being deleted.
true_branches = np.delete(true_branches, 0, 0 ) - 1
# Set the number of bits when using discrete or JVHW mutual information
# methods.
num_bits = 8
# Create a GridEst objec to prepare to run the algorithm.
node8_randPF_solar = GridEst(true_branches, vmag_matrix, 'Node8_randPF_solar',
                             num_bits)
# Choose mutual information method, either gaussian, MLE, sk_discrete, or JVHW.
mi_method = 'gaussian'
node8_randPF_solar.run_mi_alg(mi_method)
# Print successful detection rate of algorithm.
print('The SDR is: ' + str(node8_randPF_solar.find_sdr()[0]) + '%')


import numpy as np
import scipy.io as sio

poly_entro = None

def est_entro_JVHW(samp):
    """Proposed JVHW estimate of Shannon entropy (in bits) of the input sample
    This function returns a scalar JVHW estimate of the entropy of samp when
    samp is a vector, or returns a row vector containing the JVHW estimate of
    each column of samp when samp is a matrix.
    Input:
    ----- samp: a vector or matrix which can only contain integers. The input
                data type can be any interger classes such as uint8/int8/
                uint16/int16/uint32/int32/uint64/int64, or floating-point
                such as single/double.
    Output:
    ----- est: the entropy (in bits) of the input vector or that of each column
               of the input matrix. The output data type is double.
    """
    samp = formalize_sample(samp)
    [n, wid] = samp.shape
    n = float(n)

    # The order of polynomial is no more than 22 because otherwise floating-point error occurs
    order = min(4 + int(np.ceil(1.2 * np.log(n))), 22)
    global poly_entro
    if poly_entro is None:
        poly_entro = sio.loadmat('poly_coeff_entro.mat')['poly_entro']
    coeff = poly_entro[order-1, 0][0]

    f = fingerprint(samp)

    prob = np.arange(1, f.shape[0] + 1) / n

    # Piecewise linear/quadratic fit of c_1
    V1 = np.array([0.3303, 0.4679])
    V2 = np.array([-0.530556484842359, 1.09787328176926, 0.184831781602259])
    f1nonzero = f[0] > 0
    c_1 = np.zeros(wid)

    with np.errstate(divide='ignore', invalid='ignore'):
        if n >= order and f1nonzero.any():
            if n < 200:
                c_1[f1nonzero] = np.polyval(V1, np.log(n / f[0, f1nonzero]))
            else:
                n2f1_small = f1nonzero & (np.log(n / f[0]) <= 1.5)
                n2f1_large = f1nonzero & (np.log(n / f[0]) > 1.5)
                c_1[n2f1_small] = np.polyval(V2, np.log(n / f[0, n2f1_small]))
                c_1[n2f1_large] = np.polyval(V1, np.log(n / f[0, n2f1_large]))

            # make sure nonzero threshold is higher than 1/n
            c_1[f1nonzero] = np.maximum(c_1[f1nonzero], 1 / (1.9 * np.log(n)))

        prob_mat = entro_mat(prob, n, coeff, c_1)

    return np.sum(f * prob_mat, axis=0) / np.log(2)

def entro_mat(x, n, g_coeff, c_1):
    # g_coeff = {g0, g1, g2, ..., g_K}, K: the order of best polynomial approximation,
    K = len(g_coeff) - 1
    thres = 4 * c_1 * np.log(n) / n
    T, X = np.meshgrid(thres, x)
    ratio = np.minimum(np.maximum(2 * X / T - 1, 0), 1)
    q = np.arange(K).reshape((1, 1, K))
    g = g_coeff.reshape((1, 1, K + 1))
    MLE = - X * np.log(X) + 1 / (2 * n)
    polyApp = np.sum(np.concatenate((T[..., None], ((n * X)[..., None]  - q) / (T[..., None] * (n - q))), axis=2).cumprod(axis=2) * g, axis=2) - X * np.log(T)
    polyfail = np.isnan(polyApp) | np.isinf(polyApp)
    polyApp[polyfail] = MLE[polyfail]
    output = ratio * MLE + (1 - ratio) * polyApp
    return np.maximum(output, 0)


def est_entro_MLE(samp):
    """Maximum likelihood estimate of Shannon entropy (in bits) of the input
    sample
    This function returns a scalar MLE of the entropy of samp when samp is a
    vector, or returns a (row-) vector consisting of the MLE of the entropy
    of each column of samp when samp is a matrix.
    Input:
    ----- samp: a vector or matrix which can only contain integers. The input
                data type can be any interger classes such as uint8/int8/
                uint16/int16/uint32/int32/uint64/int64, or floating-point
                such as single/double.
    Output:
    ----- est: the entropy (in bits) of the input vector or that of each
               column of the input matrix. The output data type is double.
    """
    samp = formalize_sample(samp)
    [n, wid] = samp.shape
    n = float(n)

    f = fingerprint(samp)
    prob = np.arange(1, f.shape[0] + 1) / n
    prob_mat = - prob * np.log2(prob)
    return prob_mat.dot(f)


def formalize_sample(samp):
    samp = np.array(samp)
    if np.any(samp != np.fix(samp)):
        raise ValueError('Input sample must only contain integers.')
    if samp.ndim == 1 or samp.ndim == 2 and samp.shape[0] == 1:
        samp = samp.reshape((samp.size, 1))
    return samp

def fingerprint(samp):
    """A memory-efficient algorithm for computing fingerprint when wid is
    large, e.g., wid = 100
    """
    wid = samp.shape[1]

    d = np.r_[
        np.full((1, wid), True, dtype=bool),
        np.diff(np.sort(samp, axis=0), 1, 0) != 0,
        np.full((1, wid), True, dtype=bool)
    ]

    f_col = []
    f_max = 0

    for k in range(wid):
        a = np.diff(np.flatnonzero(d[:, k]))
        a_max = a.max()
        hist, _ = np.histogram(a, bins=a_max, range=(1, a_max + 1))
        f_col.append(hist)
        if a_max > f_max:
            f_max = a_max

    return np.array([np.r_[col, [0] * (f_max - len(col))] for col in f_col]).T




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 31 04:39:06 2017

@author: Daniel Speckhard, DTS@stanford.edu

This module contains methods relating to calculating the mutual information
between two distributions when we approximate the distributions as gaussian
distributions.
"""

import numpy as np

def find_gaussian_entropy(vmag_matrix):
        
    """ Calculates Gaussian entropy at a single node.
        
    This function finds the entropy by approximating the voltage
    magnitude data at different nodes as a Gaussian distribution.
        
    Args
    ----------
    self.vmag_matrix: ndarray, shape (# of data-points, # of nodes)
        This is the voltage magnitude data at different nodes and 
        timepoints.
                
    Returns
    ----------
        entropy_vec: ndarray, shape (# of nodes)
            Returns the entropy of each node with values of the entropy 
            vector (v(i)) equal to the entropy at Node i. The entropy is 
            caluclated by approximating the data at a node in the input 
            data matrix, self.vmag_matrix[:,i], as a Gaussian distribution.
    """
    num_buses = len(vmag_matrix[1,:])
    entropy_vec = np.zeros((num_buses))
    for i in range(0, num_buses):
        # We use the equation H(X) = k/2*(1+ln(2*pi)) + 1/2*ln|Sigma|.
        # Where k is the dimension of the vector X and Sigma is the
        # covariance matrix. Note, here we only have one dimension,
        # vmag(i), in our entropy calc. This is not the case if we would
        # like to include phase information into our calculation.
        #print(i)
        entropy_vec[i] = 0.5*(1+np.log(2*np.pi))+ \
                   0.5*np.log(np.var(vmag_matrix[:, i]))
        #print(entropy_vec[i])

    return entropy_vec

def find_joint_gaussian_entropy(vmag_matrix):
    """ Calculates joint Gaussian entropy between two nodes.

    This function calculates the pair-wise mutual information between
    two nodes in the input data matrix, node_volt_matrix. The mutual
    information is claculated by appoximating the distribution of data at
    each node as a Gaussian distribution.

    Args
    ----------           
    vmag_matrix: ndarray, shape (# of data-points, # of nodes)
        A matrix containing the voltage magnitude data.
        The number of columns should be the number of nodes in the system. 
        The number of rows in the matrix corresponds to the number of 
        datapoints where the voltage magnitude is measured across all 
        nodes.

    Returns
    ----------
    joint_entropy_matrix: ndarrray, shape (# of nodes, # of nodes)
        This matrix contains the joint entropy between two nodes from 
        the input matrix, self.vmag__matrix. The matrix is a lower 
        triangular matrix since the joint entropy is symmetric, i.e., 
        Entropy(i,j) = Entropy(j,i). The joint entropy of the same 
        node, Entropy(i,i), is not used in the chow-liu algorithm 
        and therefore these values (the diagonal of 
        joint_entropy_matrix) are set to zero. The entropy is 
        calculated by approximating the data at a node in the input 
        data matrix, node_volt_matrix(:,i), as a Gaussian distribution.
        We use the equation joint_entropy([X,Y]) = k/2*(1+ln(2*pi)) + 
        1/2*ln|Sigma|. Where k is the dimension of the vector [X,Y] 
        and Sigma is the covariance matrix.
        """
    num_buses = len(vmag_matrix[1,:])
    # Initialize the joint entropy matrix, note the size it determined
    # by the number of buses in the grid.
    joint_entropy_matrix = np.zeros((num_buses, num_buses))

    # We avoid calculating the joint_entropy values using the same node
    # twice. Therefore the diagonal values for the join_entropy_matrix
    # are not calculated.
    for i in range(1, num_buses):
        for k in range(0, i):

            # We use the equation joint_entropy([X,Y]) = k/2*(1+ln(2*pi)) +
            # 1/2*ln|Sigma|. Where k is the dimension of the vector [X,Y] and
            # |Sigma| is the deteriminant of the covariance matrix. For two
            # nodes, where each node contributes it's voltage magntiude data,
            # k, the dimension of [X,Y] is equal to two.

            det_of_cov_matrix = np.linalg.det(np.cov(
                    vmag_matrix[:, i], vmag_matrix[:, k]))

            # Let's check if the determinant value is very close to zero
            # and negative. If this is true, a numerical rounding error has
            # likley happened and we want to avoid getting a log of a negative
            # number.
            if (det_of_cov_matrix <= 0) and (det_of_cov_matrix > -0.0001):
                # Then we assume this is a numerical error and we make the
                # mutual information negative by setting the joint entropy
                # value to a very negative number. This is abritrary and this
                # case should never occur in practice unless we end up
                # computing the joint entropy of two nodes that are labelled
                # differently but contain the same data.
                joint_entropy_matrix[i, k] = -1E3
            else: # Othewrwise we compute as normal.
                joint_entropy_matrix[i, k] = 2/2*(
                        1+np.log(2*np.pi))+0.5*np.log(det_of_cov_matrix)
    return joint_entropy_matrix

def find_mi_mat(entropy_vec, joint_entropy_matrix):
        
    """ Finds the mutual information from between nodes using entropy vals. 
        
    The function takes in as input the single node entropy values
    (H(i)) and joint entropy matrix values (H(i,j)). The output is a lower
    triangular matrix that contains the mutual information MI(i,j). Since
    mutual information is symmetric we only have to calculate the lower
    triangular values.
        
    Args
    ----------
    Single_node_entropy_vec: This is a vector size (1 x number of
        buses) with values of the vector, v(i) equal to the entropy at
        node(i). The entropy is caluclated by approximating the data at a
        node in the input data matrix, node_volt_matrix(:,i), as a Gaussian
        distribution.
            
    joint_entropy_matrix: This matrix contains the joint entropy
        between two nodes. The matrix is a lower triangular matrix since 
        the joint entropy is symmetric, i.e., Entropy(i,j) = Entropy(j,i).
        The joint entropy of the same node, Entropy(i,j) is not used in the
        chow-liu algorithm and therefore these values (the diagonal of
        joint_entropy_matrix) are set to zero. % The entropy is caluclated 
        by approximating the data at a node in the input data matrix,
        node_volt_matrix(:,i), as a Gaussian distribution. The size of the
        matrix is (number of nodes x number of nodes). We use the
        equation joint_entropy([X,Y]) = k/2*(1+ln(2*pi)) + 1/2*ln|Sigma|. 
        Where k is the dimension of the vector [X,Y] and Sigma is the 
        covariance matrix.
            
    Returns
    ----------
    mi_matrix: This matrix is size (number of nodes x number of
        nodes). The matrix is lower triangular, meaning the diagonal and
        elements above the diagonal are zero since the mutual information
        between two nodes MI(i,j) = MI(j,i) is symmetric and the
        self-information MI(i,i) is not used in the algorithm since we 
        don't connect a node to itself (to avoid cycles). The mutual 
        information can be caluclated from the joint entropy and the single
        node entropy by recalling MI(i,j) = entropy(i) +  entropy(j) - 
        joint_entropy(i,j). """

    num_buses = len(joint_entropy_matrix[1,:])
    mi_matrix = np.zeros((num_buses, num_buses))
    for i in range(1, num_buses):
        for k in range(0, i):
            # The mutual information can be caluclated from the joint entropy
            # and the single node entropy by recalling MI(i,j) = entropy(i) +
            # entropy(j) - joint_entropy(i,j).
            mi_matrix[i, k] = entropy_vec[i] + entropy_vec[k] - \
                          joint_entropy_matrix[i, k]
    return mi_matrix